import functools
import math
import os
from typing import Callable, Optional, Union

import numpy as np
import torch
import torch.distributed as dist
import torch.nn as nn
from transformers.activations import ACT2FN

import realhf.base.constants as constants
import realhf.base.logging as logging
from realhf.impl.model.parallelism.model_parallel.modules import (
    ColumnParallelLinear,
    RowParallelLinear,
    merged_linear_with_grad_accumulation_and_async_allreduce,
)

logger = logging.getLogger("Modules")


def get_activation_fn(activation_function: str) -> Callable:
    return ACT2FN[activation_function]


SEQUENCE_PARALLEL_WARNED = False
SPLIT_KV_HEADS_WARNED = False


class LayerNormQKVLinear(nn.Module):

    def __init__(
        self,
        input_dim: int,
        head_dim: int,
        n_q_heads: int,
        n_kv_heads: int,
        layer_norm_epsilon: float,
        use_attention_bias: bool,
        layer_norm_type: Optional[str] = None,
        do_layernorm_before: bool = True,
        # parallelism
        model_parallel: bool = False,  # We set this as an option for replacing this module with layers in transformer engine
        gradient_accumulation_fusion: bool = False,
        # dtype and device
        dtype: Optional[torch.dtype] = None,
        device: Optional[torch.device] = None,
        layer_index=None,
    ):
        super().__init__()
        sequence_parallel = constants.sequence_parallel()
        if not model_parallel and (sequence_parallel or gradient_accumulation_fusion):
            global SEQUENCE_PARALLEL_WARNED
            if not SEQUENCE_PARALLEL_WARNED:
                logger.warning(
                    "sequence_parallel and gradient_accumulation_fusion are only available in model parallel mode"
                )
                SEQUENCE_PARALLEL_WARNED = True
            sequence_parallel = False
            gradient_accumulation_fusion = False
        if dtype is None:
            dtype = torch.float16
        if layer_norm_type is None:
            layer_norm_fn = nn.LayerNorm
        elif layer_norm_type == "rms":
            layer_norm_fn = LlamaRMSNorm
        elif layer_norm_type == "gemma":
            layer_norm_fn = GemmaRMSNorm
        self.ln = layer_norm_fn(
            input_dim, eps=layer_norm_epsilon, dtype=dtype, device=device
        )

        self.model_parallel = model_parallel
        self.layer_index = layer_index
        if not model_parallel:
            self.q_attn = nn.Linear(
                input_dim,
                head_dim * n_q_heads,
                bias=use_attention_bias,
                dtype=dtype,
                device=device,
            )
            self.k_attn = nn.Linear(
                input_dim,
                head_dim * n_kv_heads,
                bias=use_attention_bias,
                dtype=dtype,
                device=device,
            )
            self.v_attn = nn.Linear(
                input_dim,
                head_dim * n_kv_heads,
                bias=use_attention_bias,
                dtype=dtype,
                device=device,
            )
        else:
            self.mp_worldsize = constants.model_parallel_world_size()
            assert n_q_heads % self.mp_worldsize == 0, (
                f"n_q_heads {n_q_heads} must be divisible by "
                f"mp_worldsize {self.mp_worldsize}"
            )
            hidden_dim = input_dim
            self.q_attn = ColumnParallelLinear(
                hidden_dim,
                head_dim * n_q_heads,
                bias=use_attention_bias,
                async_tensor_model_parallel_allreduce=not sequence_parallel,
                gradient_accumulation_fusion=gradient_accumulation_fusion,
                dtype=dtype,
                device=device,
            )
            if n_kv_heads > 1 and n_kv_heads % self.mp_worldsize == 0:
                # split model parallel among heads if possible
                self.k_attn = ColumnParallelLinear(
                    hidden_dim,
                    head_dim * n_kv_heads,
                    bias=use_attention_bias,
                    async_tensor_model_parallel_allreduce=not sequence_parallel,
                    gradient_accumulation_fusion=gradient_accumulation_fusion,
                    dtype=dtype,
                    device=device,
                )
                self.v_attn = ColumnParallelLinear(
                    hidden_dim,
                    head_dim * n_kv_heads,
                    bias=use_attention_bias,
                    async_tensor_model_parallel_allreduce=not sequence_parallel,
                    gradient_accumulation_fusion=gradient_accumulation_fusion,
                    dtype=dtype,
                    device=device,
                )
            else:
                global SPLIT_KV_HEADS_WARNED
                if n_kv_heads > 1 and not SPLIT_KV_HEADS_WARNED:
                    logger.warning(
                        f"Cannot split {n_kv_heads} kv heads evenly among "
                        f"{self.mp_worldsize} model parallel ranks, "
                        f"use unsplitted linear for kv heads instead"
                    )
                    SPLIT_KV_HEADS_WARNED = True
                self.k_attn = nn.Linear(
                    hidden_dim,
                    head_dim * n_kv_heads,
                    bias=use_attention_bias,
                    dtype=dtype,
                    device=device,
                )
                self.v_attn = nn.Linear(
                    hidden_dim,
                    head_dim * n_kv_heads,
                    bias=use_attention_bias,
                    dtype=dtype,
                    device=device,
                )
                dist.all_reduce(
                    self.k_attn.weight.data,
                    op=dist.ReduceOp.SUM,
                    group=constants.model_parallel_group(),
                )
                if use_attention_bias:
                    dist.all_reduce(
                        self.k_attn.bias.data,
                        op=dist.ReduceOp.SUM,
                        group=constants.model_parallel_group(),
                    )
                dist.all_reduce(
                    self.v_attn.weight.data,
                    op=dist.ReduceOp.SUM,
                    group=constants.model_parallel_group(),
                )
                if use_attention_bias:
                    dist.all_reduce(
                        self.v_attn.bias.data,
                        op=dist.ReduceOp.SUM,
                        group=constants.model_parallel_group(),
                    )

        self.d = head_dim
        self.nq = n_q_heads
        self.nkv = n_kv_heads

        self.do_layernorm_before = do_layernorm_before

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        if self.do_layernorm_before:
            hidden_states = self.ln(x)
        if not self.model_parallel:
            q = self.q_attn(hidden_states)
            k = self.k_attn(hidden_states)
            v = self.v_attn(hidden_states)
            # qkv = self.linear(hidden_states)
            # q, k, v = torch.split(qkv, (self.d * self.nq, self.d * self.nkv, self.d * self.nkv), dim=-1)
            q = q.view(*q.shape[:-1], self.nq, self.d)
            k = k.view(*k.shape[:-1], self.nkv, self.d)
            v = v.view(*v.shape[:-1], self.nkv, self.d)
        else:
            _gradient_accumulation_fusion = self.q_attn.gradient_accumulation_fusion
            _async_grad_allreduce = self.q_attn.async_tensor_model_parallel_allreduce
            _sequence_parallel = constants.sequence_parallel()
            _is_w_parallel = [
                True,
                isinstance(self.k_attn, ColumnParallelLinear),
                isinstance(self.v_attn, ColumnParallelLinear),
            ]
            q, k, v = merged_linear_with_grad_accumulation_and_async_allreduce(
                hidden_states,
                _gradient_accumulation_fusion,
                _async_grad_allreduce,
                _sequence_parallel,
                _is_w_parallel,
                self.q_attn.weight,
                self.q_attn.bias,
                self.k_attn.weight,
                self.k_attn.bias,
                self.v_attn.weight,
                self.v_attn.bias,
            )
            q = q.view(*q.shape[:-1], self.nq // self.mp_worldsize, self.d)
            if isinstance(self.k_attn, ColumnParallelLinear):
                k = k.view(*k.shape[:-1], self.nkv // self.mp_worldsize, self.d)
                v = v.view(*v.shape[:-1], self.nkv // self.mp_worldsize, self.d)
            else:
                k = k.view(*k.shape[:-1], self.nkv, self.d)
                v = v.view(*v.shape[:-1], self.nkv, self.d)
        return q, k, v


class LayerNormMLP(nn.Module):
    SEQUENCE_PARALLEL_WARNED = False

    def __init__(
        self,
        hidden_dim: int,
        intermediate_dim: int,
        resid_pdrop: float,
        activation_function: str,
        layer_norm_epsilon: float,
        do_layernorm_before: bool = True,
        # parallelism
        model_parallel: bool = False,  # We set this as an option for replacing this module with layers in transformer engine
        gradient_accumulation_fusion: bool = False,
        # dtype and device
        dtype: Optional[torch.dtype] = None,
        device: Optional[Union[str, torch.device]] = None,
    ):
        super().__init__()
        sequence_parallel = constants.sequence_parallel()
        if not model_parallel and (sequence_parallel or gradient_accumulation_fusion):
            global SEQUENCE_PARALLEL_WARNED
            if not SEQUENCE_PARALLEL_WARNED:
                logger.warning(
                    "sequence_parallel and gradient_accumulation_fusion are only available in model parallel mode"
                )
                SEQUENCE_PARALLEL_WARNED = True
            sequence_parallel = False
            gradient_accumulation_fusion = False
        if dtype is None:
            dtype = torch.float16

        self.ln = nn.LayerNorm(
            hidden_dim, eps=layer_norm_epsilon, dtype=dtype, device=device
        )
        self.model_parallel = model_parallel
        if not model_parallel:
            self.c_fc = nn.Linear(
                hidden_dim, intermediate_dim, dtype=dtype, device=device
            )
            self.c_proj = nn.Linear(
                intermediate_dim, hidden_dim, dtype=dtype, device=device
            )
        else:
            self.c_fc = ColumnParallelLinear(
                hidden_dim,
                intermediate_dim,
                async_tensor_model_parallel_allreduce=not sequence_parallel,
                gradient_accumulation_fusion=gradient_accumulation_fusion,
                dtype=dtype,
                device=device,
            )
            self.c_proj = RowParallelLinear(
                intermediate_dim,
                hidden_dim,
                gradient_accumulation_fusion=gradient_accumulation_fusion,
                dtype=dtype,
                device=device,
            )
        self.act = get_activation_fn(activation_function)
        self.dropout = nn.Dropout(resid_pdrop)
        self.do_layernorm_before = do_layernorm_before

    def forward(self, hidden_states: torch.Tensor) -> torch.Tensor:
        if self.do_layernorm_before:
            hidden_states = self.ln(hidden_states)
        hidden_states = self.c_fc(hidden_states)
        hidden_states = self.act(hidden_states)
        hidden_states = self.c_proj(hidden_states)
        return self.dropout(hidden_states)


class LlamaLayerNormMLP(nn.Module):

    def __init__(
        self,
        hidden_dim: int,
        intermediate_dim: int,
        activation_function: str,
        layer_norm_epsilon: float,
        layer_norm_type: str,
        # parallelism
        model_parallel: bool = False,  # We set this as an option for replacing this module with layers in transformer engine
        gradient_accumulation_fusion: bool = False,
        # dtype and device
        dtype: Optional[torch.dtype] = None,
        device: Optional[Union[str, torch.device]] = None,
    ):
        super().__init__()
        sequence_parallel = constants.sequence_parallel()
        if not model_parallel and (sequence_parallel or gradient_accumulation_fusion):
            global SEQUENCE_PARALLEL_WARNED
            if not SEQUENCE_PARALLEL_WARNED:
                logger.warning(
                    "sequence_parallel and gradient_accumulation_fusion are only available in model parallel mode"
                )
                SEQUENCE_PARALLEL_WARNED = True
            sequence_parallel = False
            gradient_accumulation_fusion = False

        if dtype is None:
            dtype = torch.float16
        self.hidden_size = hidden_dim
        self.intermediate_size = intermediate_dim

        if layer_norm_type == "rms":
            self.ln = LlamaRMSNorm(
                hidden_dim, eps=layer_norm_epsilon, dtype=dtype, device=device
            )
        elif layer_norm_type == "gemma":
            self.ln = GemmaRMSNorm(
                hidden_dim, eps=layer_norm_epsilon, dtype=dtype, device=device
            )
        else:
            raise NotImplementedError()

        self.model_parallel = model_parallel
        if not model_parallel:
            self.gate_proj = nn.Linear(
                self.hidden_size,
                self.intermediate_size,
                bias=False,
                dtype=dtype,
                device=device,
            )
            self.up_proj = nn.Linear(
                self.hidden_size,
                self.intermediate_size,
                bias=False,
                dtype=dtype,
                device=device,
            )
            self.down_proj = nn.Linear(
                self.intermediate_size,
                self.hidden_size,
                bias=False,
                dtype=dtype,
                device=device,
            )
        else:
            self.gate_proj = ColumnParallelLinear(
                self.hidden_size,
                self.intermediate_size,
                async_tensor_model_parallel_allreduce=not sequence_parallel,
                gradient_accumulation_fusion=gradient_accumulation_fusion,
                bias=False,
                dtype=dtype,
                device=device,
            )
            self.up_proj = ColumnParallelLinear(
                self.hidden_size,
                self.intermediate_size,
                async_tensor_model_parallel_allreduce=not sequence_parallel,
                gradient_accumulation_fusion=gradient_accumulation_fusion,
                bias=False,
                dtype=dtype,
                device=device,
            )
            self.down_proj = RowParallelLinear(
                self.intermediate_size,
                self.hidden_size,
                gradient_accumulation_fusion=gradient_accumulation_fusion,
                bias=False,
                dtype=dtype,
                device=device,
            )
        self.act_fn = get_activation_fn(activation_function)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.ln(x)
        if not self.model_parallel:
            return self.down_proj(self.act_fn(self.gate_proj(x)) * self.up_proj(x))
        else:
            _gradient_accumulation_fusion = self.gate_proj.gradient_accumulation_fusion
            _async_grad_allreduce = self.gate_proj.async_tensor_model_parallel_allreduce
            _sequence_parallel = constants.sequence_parallel()
            _is_w_parallel = [True, True]
            gate, upproj = merged_linear_with_grad_accumulation_and_async_allreduce(
                x,
                _gradient_accumulation_fusion,
                _async_grad_allreduce,
                _sequence_parallel,
                _is_w_parallel,
                self.gate_proj.weight,
                self.gate_proj.bias,
                self.up_proj.weight,
                self.up_proj.bias,
            )
            return self.down_proj(self.act_fn(gate) * upproj)


class _LlamaRMSNorm(nn.Module):

    def __init__(
        self,
        hidden_size: int,
        eps: float = 1e-6,
        dtype: Optional[torch.dtype] = None,
        device: Optional[Union[str, torch.device]] = None,
    ):
        """
        LlamaRMSNorm is equivalent to T5LayerNorm
        """
        super().__init__()
        self.weight = nn.Parameter(torch.ones(hidden_size, dtype=dtype, device=device))
        self.variance_epsilon = eps

    def forward(self, hidden_states: torch.Tensor):
        input_dtype = hidden_states.dtype
        hidden_states = hidden_states.to(torch.float32)
        variance = hidden_states.pow(2).mean(-1, keepdim=True)
        hidden_states = hidden_states * torch.rsqrt(variance + self.variance_epsilon)
        return self.weight * hidden_states.to(input_dtype)


class GemmaRMSNorm(nn.Module):
    def __init__(
        self,
        hidden_size: int,
        eps: float = 1e-6,
        dtype: Optional[torch.dtype] = None,
        device: Optional[Union[str, torch.device]] = None,
    ):
        super().__init__()
        self.eps = eps
        self.weight = nn.Parameter(torch.ones(hidden_size, dtype=dtype, device=device))

    def _norm(self, x):
        return x * torch.rsqrt(x.pow(2).mean(-1, keepdim=True) + self.eps)

    def forward(self, x: torch.Tensor):
        output = self._norm(x.float())
        # Llama does x.to(float16) * w whilst Gemma is (x * w).to(float16)
        # See https://github.com/huggingface/transformers/pull/29402
        output = output * (1.0 + self.weight.float())
        return output.type_as(x)


if constants.use_te_impl():
    try:
        # HACK: we use transformer engine's rms norm as long as we can find the transformer engine package
        import transformer_engine.pytorch as te

        def _TELlamaRMSNorm(
            hidden_size: int,
            eps: float = 1e-6,
            dtype: Optional[torch.dtype] = None,
            device: Optional[Union[str, torch.device]] = None,
        ):
            return te.module.rmsnorm.RMSNorm(
                hidden_size=hidden_size,
                eps=eps,
                sequence_parallel=constants.sequence_parallel(),
                params_dtype=dtype,
                device=device,
            )

        LlamaRMSNorm = _TELlamaRMSNorm
    except ModuleNotFoundError:
        LlamaRMSNorm = _LlamaRMSNorm
    except ImportError:
        LlamaRMSNorm = _LlamaRMSNorm
else:
    LlamaRMSNorm = _LlamaRMSNorm

if constants.use_te_impl():
    from transformer_engine.pytorch.module.layernorm_mlp import (
        LayerNormMLP as _TELayerNormMLP,
    )

    # The same signature as LlamaLayerNormMLP
    def LlamaLayerNormMLP(
        hidden_dim: int,
        intermediate_dim: int,
        activation_function: str,
        layer_norm_epsilon: float,
        layer_norm_type: str,
        # parallelism
        model_parallel: bool = False,  # We set this as an option for replacing this module with layers in transformer engine
        gradient_accumulation_fusion: bool = False,
        # dtype and device
        dtype: Optional[torch.dtype] = None,
        device: Optional[Union[str, torch.device]] = None,
    ):
        assert layer_norm_type == "rms"
        assert activation_function == "silu"
        return _TELayerNormMLP(
            hidden_size=hidden_dim,
            ffn_hidden_size=intermediate_dim,
            eps=layer_norm_epsilon,
            sequence_parallel=constants.sequence_parallel(),
            return_bias=False,
            tp_group=constants.model_parallel_group(),
            tp_size=constants.model_parallel_world_size(),
            bias=False,
            normalization="RMSNorm",
            activation="swiglu",
            fuse_wgrad_accumulation=gradient_accumulation_fusion,
            params_dtype=dtype,
            set_parallel_mode=model_parallel,
            device=device,
        )



import functools
import math
from typing import Callable, Optional, Union

import numpy as np
import torch
import torch.distributed as dist
import torch.nn as nn

import realhf.base.logging as logging
from realhf.base.constants import data_parallel_group

logger = logging.getLogger("Modules")


class ExponentialRunningMeanStd(nn.Module):

    def __init__(self, beta=0.999, epsilon=1e-5, high_precision=True):
        super().__init__()
        self.__beta = beta
        self.__eps = epsilon

        self.__dtype = torch.float64 if high_precision else torch.float32

        self.__mean = nn.Parameter(
            torch.zeros((1,), dtype=self.__dtype, device="cuda"),
            requires_grad=False,
        )
        self.__mean_sq = nn.Parameter(
            torch.zeros((1,), dtype=self.__dtype, device="cuda"),
            requires_grad=False,
        )
        self.__debiasing_term = nn.Parameter(
            torch.zeros((1,), dtype=self.__dtype, device="cuda"),
            requires_grad=False,
        )

        self.reset_parameters()

    def reset_parameters(self):
        self.__mean.zero_()
        self.__mean_sq.zero_()
        self.__debiasing_term.zero_()

    def forward(self, *args, **kwargs):
        # we don't implement the forward function because its meaning
        # is somewhat ambiguous
        raise NotImplementedError()

    @torch.no_grad()
    def update(self, x: torch.Tensor, mask: Optional[torch.Tensor] = None):
        x = x.to(self.__dtype)
        if mask is not None:
            mask = mask.to(self.__dtype)
        if mask is None:
            factor = torch.tensor(np.prod(x.shape), dtype=self.__dtype, device=x.device)
        else:
            x = x * mask
            factor = mask.sum()

        x_sum = x.sum()
        x_sum_sq = x.square().sum()
        if dist.is_initialized():
            dist.all_reduce(factor, op=dist.ReduceOp.SUM, group=data_parallel_group())
            dist.all_reduce(x_sum, op=dist.ReduceOp.SUM, group=data_parallel_group())
            dist.all_reduce(x_sum_sq, op=dist.ReduceOp.SUM, group=data_parallel_group())
        batch_mean = x_sum / factor
        batch_sq_mean = x_sum_sq / factor

        self.__mean.data[:] = self.__beta * self.__mean.data[:] + batch_mean * (
            1.0 - self.__beta
        )
        self.__mean_sq.data[:] = self.__beta * self.__mean_sq.data[
            :
        ] + batch_sq_mean * (1.0 - self.__beta)
        self.__debiasing_term.data[:] = (
            self.__beta * self.__debiasing_term.data[:] + 1.0 - self.__beta
        )

    @torch.no_grad()
    def mean_std(self):
        debiased_mean = self.__mean / self.__debiasing_term.clamp(min=self.__eps)
        debiased_mean_sq = self.__mean_sq / self.__debiasing_term.clamp(min=self.__eps)
        debiased_var = (debiased_mean_sq - debiased_mean**2).clamp(min=1e-2)
        return debiased_mean, debiased_var.sqrt()

    @torch.no_grad()
    def normalize(self, x):
        x = x.to(self.__dtype)
        mean, std = self.mean_std()
        return (
            ((x - mean) / std).clip(-5, 5).float()
        )  # clipping is a trick from hide and seek

    @torch.no_grad()
    def denormalize(self, x):
        x = x.to(self.__dtype)
        mean, std = self.mean_std()
        return (x * std + mean).float()


class MovingAverageRunningMeanStd(nn.Module):

    def __init__(self, high_precision=True):
        super().__init__()

        self.__dtype = torch.float64 if high_precision else torch.float32

        self.__mean = nn.Parameter(
            torch.zeros((1,), dtype=self.__dtype, device="cuda"),
            requires_grad=False,
        )
        self.__mean_sq = nn.Parameter(
            torch.zeros((1,), dtype=self.__dtype, device="cuda"),
            requires_grad=False,
        )
        self.__accum_denominator = 0

        self.reset_parameters()

    def reset_parameters(self):
        self.__mean.zero_()
        self.__mean_sq.zero_()
        self.__accum_denominator = 0

    def forward(self, *args, **kwargs):
        # we don't implement the forward function because its meaning
        # is somewhat ambiguous
        raise NotImplementedError()

    @torch.no_grad()
    def update(self, x: torch.Tensor, mask: Optional[torch.Tensor] = None):
        x = x.to(self.__dtype)
        if mask is not None:
            mask = mask.to(self.__dtype)
        if mask is None:
            factor = torch.tensor(np.prod(x.shape), dtype=self.__dtype, device=x.device)
        else:
            x = x * mask
            factor = mask.sum()

        x_sum = x.sum()
        x_sum_sq = x.square().sum()
        if dist.is_initialized():
            dist.all_reduce(factor, op=dist.ReduceOp.SUM, group=data_parallel_group())
            dist.all_reduce(x_sum, op=dist.ReduceOp.SUM, group=data_parallel_group())
            dist.all_reduce(x_sum_sq, op=dist.ReduceOp.SUM, group=data_parallel_group())

        self.__mean.data[:] = (
            self.__accum_denominator * self.__mean.data[:] + x_sum
        ) / (self.__accum_denominator + factor)
        self.__mean_sq.data[:] = (
            self.__accum_denominator * self.__mean_sq.data[:] + x_sum_sq
        ) / (self.__accum_denominator + factor)
        self.__accum_denominator += factor

    @torch.no_grad()
    def mean_std(self):
        return (
            self.__mean.clone(),
            (self.__mean_sq - self.__mean**2).clamp(min=1e-2).sqrt(),
        )

    @torch.no_grad()
    def normalize(self, x):
        x = x.to(self.__dtype)
        mean, std = self.mean_std()
        return (
            ((x - mean) / std).clip(-5, 5).float()
        )  # clipping is a trick from hide and seek

    @torch.no_grad()
    def denormalize(self, x):
        x = x.to(self.__dtype)
        mean, std = self.mean_std()
        return (x * std + mean).float()


class PopArtValueHead(nn.Module):

    def __init__(
        self,
        input_dim,
        critic_dim,
        beta=0.99999,
        epsilon=1e-5,
        burn_in_updates=torch.inf,
        rms_type: str = "exp",
        high_precision=True,
    ):
        super().__init__()
        if rms_type == "exp":
            rms_cls = functools.partial(ExponentialRunningMeanStd, beta=beta)
        elif rms_type == "ma":
            rms_cls = MovingAverageRunningMeanStd
        else:
            raise NotImplementedError(f"Unknown rms type {rms_type}")
        self.__rms = rms_cls(
            input_shape=(critic_dim,),
            epsilon=epsilon,
            high_precision=high_precision,
        )

        self.__weight = nn.Parameter(torch.zeros(critic_dim, input_dim))
        self.__bias = nn.Parameter(torch.zeros(critic_dim))
        # The same initialization as `nn.Linear`.
        torch.nn.init.kaiming_uniform_(self.__weight, a=math.sqrt(5))
        torch.nn.init.uniform_(
            self.__bias, -1 / math.sqrt(input_dim), 1 / math.sqrt(input_dim)
        )

        self.__burn_in_updates = burn_in_updates
        self.__update_cnt = 0

    @property
    def weight(self):
        return self.__weight

    @property
    def bias(self):
        return self.__bias

    def forward(self, feature):
        return torch.nn.functional.linear(feature, self.__weight, self.__bias)

    @torch.no_grad()
    def update(self, x, mask):
        old_mean, old_std = self.__rms.mean_std()
        self.__rms.update(x, mask)
        new_mean, new_std = self.__rms.mean_std()
        self.__update_cnt += 1

        if self.__update_cnt > self.__burn_in_updates:
            self.__weight.data[:] = self.__weight * (old_std / new_std).unsqueeze(-1)
            self.__bias.data[:] = (
                old_std * self.__bias + old_mean - new_mean
            ) / new_std

    @torch.no_grad()
    def normalize(self, x):
        return self.__rms.normalize(x)

    @torch.no_grad()
    def denormalize(self, x):
        return self.__rms.denormalize(x)



import dataclasses
import os
import pickle
import time
from typing import *

import pytest
import torch
import torch.distributed as dist
import torch.utils.data
import transformers

import realhf.api.core.model_api as model_api
from realhf.api.quickstart.model import ParallelismConfig
from realhf.base import constants, logging, testing

logger = logging.getLogger("tests.test_generate")

# NOTE: To run test for a new model class, add a new entry to MODEL_CLASS_TO_PATH
# with the model class name as the key and the path to the model weights as the value.
MODEL_CLASS_TO_PATH = {
    "llama": "/lustre/public/pretrained_model_weights/Llama-2-7b-hf",
}
_available_model_classes = []
for k, v in MODEL_CLASS_TO_PATH.items():
    if os.path.exists(v):
        _available_model_classes.append(k)


@pytest.fixture(params=_available_model_classes)
def model_class(request):
    return request.param


@dataclasses.dataclass
class GenerateTestParams:
    parallel: ParallelismConfig = dataclasses.field(default_factory=ParallelismConfig)
    huggingface: bool = False
    prompt_len: int = 64
    gen_len: int = 64
    batch_size: int = 32
    n_batches: int = 2
    use_cuda_graph: bool = False

    def identifier(self):
        s = "real" if not self.huggingface else "huggingface"
        s += f"dp{self.parallel.data_parallel_size}"
        s += f"mp{self.parallel.model_parallel_size}"
        s += f"pp{self.parallel.pipeline_parallel_size}"
        s += f"G" if self.use_cuda_graph else ""
        return s


def check_sequences_consistency(
    batched_seq1: torch.LongTensor, batched_seq2: torch.LongTensor
):
    matched_tokens = 0
    matched_seqs = 0
    total_tokens = 0
    assert len(batched_seq1) == len(batched_seq2)
    for i in range(len(batched_seq1)):
        a = batched_seq1[i]
        b = batched_seq2[i]
        assert torch.is_tensor(a) and torch.is_tensor(b)
        assert a.dim() == 1 and b.dim() == 1, (a.shape, b.shape)
        gen_len = a.shape[0] if a.shape[0] < b.shape[0] else b.shape[0]
        b = b[:gen_len]
        a = a[:gen_len]
        for j in range(gen_len):
            if a[j] != b[j]:
                logger.info(f"Mismatch at sequence {i} position {j}")
                break
            matched_tokens += 1
        else:
            matched_seqs += 1
        total_tokens += gen_len
    logger.info(
        f"Matched {matched_seqs}/{len(batched_seq1)} "
        f"sequences and {matched_tokens}/{total_tokens} tokens"
    )
    return (
        matched_seqs,
        matched_tokens,
        float(matched_tokens) / total_tokens,
        float(matched_seqs) / len(batched_seq1),
    )


def check_generation_consistency(
    model_class: str,
    save_result_path: str,
    test_params_list: List[GenerateTestParams],
    seq_acc_threshold: float = 1.0,
    token_acc_threshold: float = 1.0,
):
    identifiers = [test_params.identifier() for test_params in test_params_list]
    dp_rank_counter = {identifier: 0 for identifier in identifiers}
    for f in os.listdir(save_result_path):
        if not f.endswith(".pkl"):
            continue
        identifier, _, _ = f.split("-")
        if identifier in identifiers:
            dp_rank_counter[identifier] += 1

    results = {}
    for identifier in identifiers:
        tmp = []
        for i in range(dp_rank_counter[identifier]):
            fn = f"{identifier}-{model_class}-{i}.pkl"
            logger.info(f"loading file {fn}")
            load_path = os.path.join(save_result_path, fn)
            t = pickle.load(open(load_path, "rb"))
            tmp.append(t)
        res = torch.cat(tmp, dim=0)
        logger.info(f"test {identifier} loaded result shape {res.shape}")
        results[identifier] = res

    baseline_result = results[identifiers[0]]
    for identifier, result in results.items():
        if identifier == identifiers[0]:
            continue
        _, _, seq_acc, token_acc = check_sequences_consistency(baseline_result, result)
        assert seq_acc >= seq_acc_threshold and token_acc >= token_acc_threshold, (
            identifier,
            identifiers[0],
            seq_acc,
            token_acc,
        )


def save_result_filename(
    model_class: str,
    save_result_path: str,
    dp_rank: int,
    test_params: GenerateTestParams,
):
    fn = test_params.identifier()
    fn += f"-{model_class}"
    fn += f"-{dp_rank}.pkl"
    return os.path.join(save_result_path, fn)


@pytest.fixture
def model_path(model_class: str):
    return MODEL_CLASS_TO_PATH[model_class]


@pytest.fixture
def save_result_path(tmpdir_factory: pytest.TempdirFactory):
    return tmpdir_factory.mktemp("save_result")


@pytest.fixture
def tokenizer(model_path: str):
    return model_api.load_hf_tokenizer(model_path)


def make_random_inference_model(model_class, model_path):
    with constants.model_scope(testing.MODEL_NAME):
        from realhf.impl.model.backend.inference import PipelinableInferenceEngine
        from realhf.impl.model.nn.real_llm_api import ReaLModel, add_helper_functions

        mconfig = getattr(ReaLModel, f"config_from_{model_class}")(
            model_path=model_path,
            is_critic=False,
        )
        model = ReaLModel(mconfig, dtype=torch.float16, device="cuda")
        if constants.pipe_parallel_world_size() == 1:
            model = add_helper_functions(model)
        model.instantiate()
        getattr(model, f"from_{model_class}")(
            load_dir=model_path, init_critic_from_actor=False
        )
        model = PipelinableInferenceEngine(model)
        model.eval()
    return model


def make_huggingface_inference_model(model_path):
    hf_model = transformers.AutoModelForCausalLM.from_pretrained(model_path).to(
        dtype=torch.float16, device="cuda"
    )
    hf_model.eval()
    return hf_model


@torch.no_grad()
def real_model_generate(
    model_class: str,
    tokenizer: transformers.PreTrainedTokenizer,
    model_path: str,
    save_result_path: str,
    test_params: GenerateTestParams,
):
    from realhf.api.core.model_api import GenerationHyperparameters
    from realhf.base.namedarray import NamedArray

    assert not test_params.huggingface
    parallel = test_params.parallel
    generation_config = GenerationHyperparameters(
        min_new_tokens=test_params.gen_len,
        max_new_tokens=test_params.gen_len,
        use_cuda_graph=test_params.use_cuda_graph,
        greedy=True,
    )
    testing.init_global_constants(
        num_dp=parallel.data_parallel_size,
        num_mp=parallel.model_parallel_size,
        num_pp=parallel.pipeline_parallel_size,
        sequence_parallel=parallel.use_sequence_parallel,
        max_prompt_len=test_params.prompt_len + test_params.gen_len,
    )
    model = make_random_inference_model(model_class, model_path)
    with constants.model_scope(testing.MODEL_NAME):
        data_batches = testing.make_random_packed_batches(
            test_params.n_batches,
            test_params.batch_size,
            test_params.prompt_len,
            tokenizer.vocab_size,
            seed=1,
        )
        if_save_result = (
            constants.pipe_parallel_rank() == constants.pipe_parallel_world_size() - 1
            and constants.model_parallel_rank() == 0
        )
        results = []
        for i, data_batch in enumerate(data_batches):
            data_batch: NamedArray
            packed_prompts = data_batch["packed_prompts"].cuda()
            prompt_lengths = torch.full(
                (test_params.batch_size // constants.data_parallel_world_size(),),
                test_params.prompt_len,
                dtype=torch.int32,
            ).cuda()
            prompt_cu_seqlens = torch.nn.functional.pad(
                prompt_lengths.cumsum(0), (1, 0)
            )
            logger.info(
                f"Batch {i} rank {constants.parallelism_rank()} "
                f"prompt shape {packed_prompts.shape}"
            )
            st = time.monotonic()
            res = model.generate(
                seqlens_cpu=data_batch.metadata["seqlens"],
                tokenizer=tokenizer,
                packed_input_ids=packed_prompts,
                cu_seqlens=prompt_cu_seqlens,
                gconfig=generation_config,
            )
            t = time.monotonic() - st

            if res is None:
                logger.info(f"Batch {i} rank {dist.get_rank()} return None")
            else:
                gen_tokens, logprobs, logits_mask, *_ = res
                logger.info(
                    f"Batch {i} rank {dist.get_rank()} generated shape = {gen_tokens.shape}, "
                    f"logprobs shape = {logprobs.shape}"
                )
                if if_save_result and res is not None:
                    results.append(gen_tokens)
            logger.info(f"Batch {i} rank {dist.get_rank()} time = {t:.2f}s")

        if if_save_result and len(results) > 0:
            results = torch.cat(results, dim=0)
            path = save_result_filename(
                model_class,
                save_result_path,
                constants.data_parallel_rank(),
                test_params,
            )
            pickle.dump(results, open(path, "wb"))


def huggingface_model_generate(
    model_class: str,
    tokenizer: transformers.PreTrainedTokenizer,
    model_path: str,
    save_result_path: str,
    test_params: GenerateTestParams,
):
    assert test_params.huggingface
    parallel = test_params.parallel
    assert (
        parallel.data_parallel_size
        == parallel.model_parallel_size
        == parallel.pipeline_parallel_size
        == 1
    )
    data_batches = testing.make_random_unpacked_batches(
        test_params.n_batches,
        test_params.batch_size,
        test_params.prompt_len,
        tokenizer.vocab_size,
        seed=1,
        dp_rank=0,
        dp_size=1,
    )
    model = make_huggingface_inference_model(model_path)
    results = []

    for i, batch in enumerate(data_batches):
        input_ids = batch["input_ids"].cuda()
        attention_mask = batch["attention_mask"].cuda()
        prompt_len = input_ids.shape[1]
        res = model.generate(
            input_ids=input_ids,
            attention_mask=attention_mask,
            max_new_tokens=test_params.gen_len,  # match generate output length
            min_new_tokens=test_params.gen_len,
            do_sample=False,
        )
        res = res[:, prompt_len:]
        results.append(res)

    if len(results) > 0:
        results = torch.cat(results, dim=0)
        path = save_result_filename(model_class, save_result_path, 0, test_params)
        pickle.dump(results, open(path, "wb"))


real_simple = GenerateTestParams()
huggingface = GenerateTestParams(huggingface=True)
real_cudagraph = GenerateTestParams(use_cuda_graph=True)
real_3d_parallel = GenerateTestParams(parallel=ParallelismConfig(2, 2, 2, False))
real_3d_parallel_cudagraph = GenerateTestParams(
    parallel=ParallelismConfig(2, 2, 2, False), use_cuda_graph=True
)
real_mp = GenerateTestParams(parallel=ParallelismConfig(8, 1, 1, False))
real_mp_cudagraph = GenerateTestParams(
    parallel=ParallelismConfig(8, 1, 1, False), use_cuda_graph=True
)
real_dp = GenerateTestParams(parallel=ParallelismConfig(1, 1, 8, False))
real_dp_cudagraph = GenerateTestParams(
    parallel=ParallelismConfig(1, 1, 8, False), use_cuda_graph=True
)
real_pp = GenerateTestParams(parallel=ParallelismConfig(1, 4, 1, False))
real_pp_cudagraph = GenerateTestParams(
    parallel=ParallelismConfig(1, 4, 1, False), use_cuda_graph=True
)


@pytest.mark.parametrize(
    "test_params1,test_params2,seq_acc_threshold,token_acc_threshold",
    [
        (real_simple, huggingface, 0.8, 0.8),
        (real_simple, real_cudagraph, 1.0, 1.0),
        (real_simple, real_3d_parallel, 0.8, 0.8),
        (real_mp, real_mp_cudagraph, 0.8, 0.8),
        (real_dp, real_dp_cudagraph, 1.0, 1.0),
        (real_pp, real_pp_cudagraph, 1.0, 1.0),
        (real_3d_parallel, real_3d_parallel_cudagraph, 0.8, 0.8),
    ],
)
@pytest.mark.slow
@pytest.mark.distributed
@pytest.mark.skipif(
    not torch.cuda.is_available() or len(_available_model_classes) == 0,
    reason="either no GPUs or no model weights found",
)
def test_generate_consistency(
    model_class: str,
    tokenizer: Any,
    model_path: str,
    save_result_path: str,
    test_params1: GenerateTestParams,
    test_params2: GenerateTestParams,
    seq_acc_threshold: float,
    token_acc_threshold: float,
):
    def launch_test(test_params: GenerateTestParams):
        func = (
            huggingface_model_generate
            if test_params.huggingface
            else real_model_generate
        )
        ngpus = (
            test_params.parallel.data_parallel_size
            * test_params.parallel.model_parallel_size
            * test_params.parallel.pipeline_parallel_size
        )
        assert ngpus <= torch.cuda.device_count()

        test = testing.LocalMultiProcessTest(
            ngpus,
            func,
            model_class,
            tokenizer,
            model_path,
            save_result_path,
            test_params,
        )
        test.launch()

    launch_test(test_params1)
    launch_test(test_params2)

    check_generation_consistency(
        model_class,
        save_result_path,
        [test_params1, test_params2],
        seq_acc_threshold,
        token_acc_threshold,
    )



import contextlib
import dataclasses
import math
import os
import subprocess
from typing import *

import numpy as np
import torch
import torch.utils.cpp_extension as torch_cpp_ext
from packaging.version import Version, parse

import realhf
from realhf.api.core import model_api
from realhf.api.core.config import ModelName
from realhf.base import logging

from .real_llm_base import (
    real_model_embedding_param_keys,
    real_model_head_param_keys,
    real_model_tblock_param_keys,
)
from .real_llm_parallel import (
    get_real_model_param_shape,
    intervals_partition_fn,
    mp_partition_key,
    shape_partition_fn,
)

try:
    from realhf._C.interval_op import merge_intervals
except ImportError:
    merge_intervals = None
logger = logging.getLogger("FlattenParam")

_FLAT_PARAM_INDICES_CACHE = {}
_SET_INTERVAL_FN = {}
_GET_INTERVAL_FN = {}


def get_nvcc_cuda_version(cuda_dir: str) -> Version:
    """Get the CUDA version from nvcc.

    Adapted from https://github.com/NVIDIA/apex/blob/8b7a1ff183741dd8f9b87e7bafd04cfde99cea28/setup.py
    """
    nvcc_output = subprocess.check_output(
        [cuda_dir + "/bin/nvcc", "-V"], universal_newlines=True
    )
    output = nvcc_output.split()
    release_idx = output.index("release") + 1
    nvcc_cuda_version = parse(output[release_idx].split(",")[0])
    return nvcc_cuda_version


def _torch_dtype_to_cpp_dtype(dtype: torch.dtype) -> str:
    if dtype == torch.float32:
        return "float"
    if dtype == torch.float64:
        return "double"
    if dtype == torch.float16:
        return "at::Half"
    if dtype == torch.bfloat16:
        return "at::BFloat16"
    raise ValueError(f"Unsupported dtype: {dtype}")


@dataclasses.dataclass
class ContiguousParamSpec:
    start_idx: int
    end_idx: int
    shape: torch.Size


def _is_integer_list_contiguous(l: List[int]) -> bool:
    return np.all(np.array(l) == np.arange(len(l)) + l[0])


def _are_intervals_contiguous(l: List[Tuple[int, int]]) -> bool:
    l = sorted(l, key=lambda x: x[0])
    res = True
    for i in range(len(l) - 1):
        res &= l[i][1] == l[i + 1][0]
    return res


def recursive_getattr(obj, attr_string):
    attrs = attr_string.split(".")
    for attr in attrs:
        obj = getattr(obj, attr)
    return obj


def _slice_interval_dispatch(N: int) -> Callable:
    global _GET_INTERVAL_FN
    if N in _GET_INTERVAL_FN:
        return _GET_INTERVAL_FN[N]
    src_path = os.path.join(
        realhf.__path__[0], os.pardir, "csrc", "interval_op", "slice_interval.cpp"
    )
    with open(src_path, "r") as f:
        src = f.read()
    # NOTE: We embed the C++ binding code in python docstring because we want to dynamically compile
    # for a specific template value N. This ensures the optimal performance.
    src += f"""
PYBIND11_MODULE(TORCH_EXTENSION_NAME, m) {{
  m.def("slice_intervals", &slice_intervals<{N}>, "Slice intervals of a 1D tensor");
}}
"""
    ABI = 1 if torch._C._GLIBCXX_USE_CXX11_ABI else 0
    logger.info(f">>> Building slice interval extension with N={N}. <<<")
    slice_fn = torch_cpp_ext.load_inline(
        f"slice_intervals_{N}",
        cpp_sources=[src],
        extra_cflags=[
            "-O3",
            "-std=c++17",
            f"-D_GLIBCXX_USE_CXX11_ABI={ABI}",
            "-fopenmp",
        ],
        extra_ldflags=["-fopenmp"],
        verbose=False,
    ).slice_intervals
    _GET_INTERVAL_FN[N] = slice_fn
    return slice_fn


def _set_interval_dispatch(N: int, torch_dtype: torch.dtype) -> Callable:
    global _SET_INTERVAL_FN
    dtype = _torch_dtype_to_cpp_dtype(torch_dtype)
    if (N, dtype) in _SET_INTERVAL_FN:
        return _SET_INTERVAL_FN[(N, dtype)]

    src_path = os.path.join(
        realhf.__path__[0], os.pardir, "csrc", "interval_op", "set_interval.cu"
    )
    with open(src_path, "r") as f:
        src = f.read()
    # NOTE: We embed the C++ binding code in python docstring because we want to dynamically compile
    # for a specific template value N. This ensures the optimal performance.
    src += f"""
PYBIND11_MODULE(TORCH_EXTENSION_NAME, m) {{
  m.def("set_intervals", &set_intervals<{N}, {dtype}>, "Set intervals of a 1D tensor");
}}
"""
    ABI = 1 if torch._C._GLIBCXX_USE_CXX11_ABI else 0
    NVCC_FLAGS = ["-O3", "-std=c++17", f"-D_GLIBCXX_USE_CXX11_ABI={ABI}"]
    nvcc_cuda_version = get_nvcc_cuda_version(torch_cpp_ext.CUDA_HOME)
    # Use NVCC threads to parallelize the build.
    if nvcc_cuda_version >= Version("11.2"):
        nvcc_threads = int(os.getenv("NVCC_THREADS", 8))
        num_threads = min(os.cpu_count(), nvcc_threads)
        NVCC_FLAGS += ["--threads", str(num_threads)]

    if nvcc_cuda_version >= Version("11.8"):
        NVCC_FLAGS += ["-DENABLE_FP8_E5M2"]
    NVCC_FLAGS += torch_cpp_ext.COMMON_NVCC_FLAGS
    REMOVE_NVCC_FLAGS = [
        "-D__CUDA_NO_HALF_OPERATORS__",
        "-D__CUDA_NO_HALF_CONVERSIONS__",
        "-D__CUDA_NO_BFLOAT16_CONVERSIONS__",
        "-D__CUDA_NO_HALF2_OPERATORS__",
    ]
    for flag in REMOVE_NVCC_FLAGS:
        with contextlib.suppress(ValueError):
            torch_cpp_ext.COMMON_NVCC_FLAGS.remove(flag)
    logger.info(
        f">>> Building set interval extension with N={N} and dtype={dtype}. <<<"
    )
    set_fn = torch_cpp_ext.load_inline(
        f"set_intervals_{N}_{str(torch_dtype).split('.')[1]}",
        cpp_sources=[],
        cuda_sources=[src],
        extra_cuda_cflags=NVCC_FLAGS,
        with_cuda=True,
        verbose=False,
    ).set_intervals
    _SET_INTERVAL_FN[(N, dtype)] = set_fn
    return set_fn


def _slice_intervals_py(src: torch.Tensor, intervals: List[Tuple[int, int]]):
    # Drop-in replacement for the C++ implementation.
    assert len(src.shape) == 1
    assert all([x[0] >= 0 for x in intervals])
    assert all([x[1] <= src.shape[0] for x in intervals])
    N = len(intervals)
    slices = []
    for i, j in intervals:
        slices.append(src[i:j])
    return torch.cat(slices, dim=0)


def _set_intervals_py(
    src: torch.Tensor,
    dst: torch.Tensor,
    intervals: List[Tuple[int, int]],
):
    # Drop-in replacement for the C++ implementation.
    assert len(dst.shape) == len(src.shape) == 1
    offset = 0
    for i, j in intervals:
        assert i >= 0
        assert j <= dst.shape[0], (j, dst.shape[0])
        dst[i:j] = src[offset : offset + j - i]
        offset += j - i
    assert offset == src.shape[0]


def slice_intervals(
    src: torch.Tensor,
    intervals: List[Tuple[int, int]],
):
    N = len(intervals)
    if N < 1000:
        # NOTE: The C++ implementation has no negative effect, but for small Ns
        # using the python implementation can omit compliation.
        return _slice_intervals_py(src, intervals)
    slice_fn = _slice_interval_dispatch(N)
    return slice_fn(src, intervals)


def set_intervals(
    src: torch.Tensor,
    dst: torch.Tensor,
    intervals: List[Tuple[int, int]],
):
    if len(intervals) < 2048 or not (src.is_cuda and dst.is_cuda):
        # NOTE: The CUDA implementation will launch a thread for each interval,
        # which has a negative effect when the number of intervals is small.
        return _set_intervals_py(src, dst, intervals)
    set_fn = _set_interval_dispatch(len(intervals), src.dtype)
    return set_fn(src, dst, intervals)


def param_size_from_keys(
    config: model_api.ReaLModelConfig,
    src_mp_size: int,
    sequence_parallel: bool,
    sd_keys: List[str],
    src2dst_tp_size: int,
    src2dst_tp_rank: int,
    head_param_point_to_embedding: bool,
) -> Tuple[List[int], int]:
    param_size = 0
    for k in sd_keys:
        if (
            head_param_point_to_embedding
            and k == f"{config.n_layers + 1}.weight"
            and "0.wte.weight" in sd_keys
        ):
            continue
        new_shape = mp_partition_key(
            k,
            get_real_model_param_shape(k, config, src_mp_size, sequence_parallel),
            src2dst_tp_rank,
            src2dst_tp_size,
            config,
            partition_fn=shape_partition_fn,
        )
        param_size += int(np.prod(new_shape))
    return param_size


def build_param_spec(
    layer_indices: List[int],
    config: model_api.ReaLModelConfig,
    dp_size: int,
    mp_size: int,
    pp_size: int,
    sequence_parallel: bool,
    head_param_point_to_embedding: bool,
    bucket_size: int = 40000000,
) -> Tuple[Dict[str, ContiguousParamSpec], int]:
    # TODO: omit parameters that do not require gradient?
    if len(layer_indices) == 0:
        return {}, 0

    disable_bucketing = 0 not in layer_indices

    sd_keys = []
    for layer_idx in sorted(layer_indices):
        if layer_idx == 0:
            sd_keys += real_model_embedding_param_keys(config)
        elif layer_idx == config.n_layers + 1:
            sd_keys += real_model_head_param_keys(config)
        else:
            sd_keys += real_model_tblock_param_keys(config, layer_idx - 1)

    # In the reverse order as backpropagation, consistent with Megatron.
    sd_keys = list(reversed(sd_keys))

    data_start_index = 0
    bucket_data_start_index = data_start_index
    bucket_params = set()

    def _requires_new_allreduce_bucket(k):
        if pp_size == 1:
            return False
        if config.is_critic:
            return False
        if not config.tied_embedding:
            return False
        return k == f"{config.n_layers + 1}.weight" or k == "0.wte.weight"

    def _pad_to_multiple(x, m):
        return int(math.ceil(x / m)) * m

    def _create_fake_bucket(data_end_index) -> int:
        nonlocal bucket_data_start_index, bucket_params
        data_end_index = _pad_to_multiple(data_end_index, dp_size)
        # Update bucket metadata.
        bucket_data_start_index = data_end_index
        # Re-set bucket_params and increment bucket_id for next bucket.
        bucket_params = set()
        # Return the potentially padded data_end_index.
        return data_end_index

    param_spec = {}
    for k in sd_keys:
        if head_param_point_to_embedding and k == f"{config.n_layers + 1}.weight":
            continue

        shape = get_real_model_param_shape(k, config, mp_size, sequence_parallel)
        numel = int(np.prod(shape))
        data_end_index = data_start_index + numel

        if _requires_new_allreduce_bucket(k) and len(bucket_params) > 0:
            _create_fake_bucket(data_start_index)

        param_spec[k] = ContiguousParamSpec(
            data_start_index,
            data_end_index,
            shape,
        )
        bucket_params.add(k)
        if (
            not disable_bucketing
            and (data_end_index - bucket_data_start_index) >= bucket_size
        ) or _requires_new_allreduce_bucket(k):
            data_end_index = _create_fake_bucket(data_end_index)

        data_start_index = data_end_index

    if len(bucket_params) > 0:
        data_end_index = _create_fake_bucket(data_end_index)

    if head_param_point_to_embedding and f"{config.n_layers + 1}.weight" in sd_keys:
        param_spec[f"{config.n_layers + 1}.weight"] = param_spec["0.wte.weight"]
    return param_spec, data_end_index


def param_intervals_from_keys(
    model_name: ModelName,
    config: model_api.ReaLModelConfig,
    head_param_point_to_embedding: bool,
    param_spec: Dict[str, ContiguousParamSpec],
    mp_size: int,
    sequence_parallel: bool,
    sd_keys: List[str],
    portion_size: int,
    portion_rank: int,
) -> List[int]:
    param_size = param_size_from_keys(
        config=config,
        src_mp_size=mp_size,
        sequence_parallel=sequence_parallel,
        sd_keys=sd_keys,
        src2dst_tp_size=portion_size,
        src2dst_tp_rank=portion_rank,
        head_param_point_to_embedding=head_param_point_to_embedding,
    )

    interval_size = 0
    intervals = []
    for k in sd_keys:
        if (
            head_param_point_to_embedding
            and k == f"{config.n_layers + 1}.weight"
            and "0.wte.weight" in sd_keys
        ):
            continue
        if (
            model_name,
            k.split(".", 1)[1],
            mp_size,
            portion_rank,
            portion_size,
        ) not in _FLAT_PARAM_INDICES_CACHE:
            zero_start_intervals = mp_partition_key(
                k,
                get_real_model_param_shape(k, config, mp_size, sequence_parallel),
                portion_rank,
                portion_size,
                config,
                partition_fn=intervals_partition_fn,
            )
            _FLAT_PARAM_INDICES_CACHE[
                (
                    model_name,
                    k.split(".", 1)[1],
                    mp_size,
                    portion_rank,
                    portion_size,
                )
            ] = zero_start_intervals
        else:
            zero_start_intervals = _FLAT_PARAM_INDICES_CACHE[
                (
                    model_name,
                    k.split(".", 1)[1],
                    mp_size,
                    portion_rank,
                    portion_size,
                )
            ]
        intervals += (zero_start_intervals + param_spec[k].start_idx).tolist()
        interval_size += sum(zero_start_intervals[:, 1] - zero_start_intervals[:, 0])
    # assert len(set([x[0] for x in intervals])) == len(intervals)
    assert interval_size == param_size, (interval_size, param_size)
    intervals = sorted(intervals, key=lambda x: x[0])
    if merge_intervals is not None:
        intervals = merge_intervals(intervals)
    return intervals


def map_param_to_contigous_memory(
    layers: torch.nn.ModuleList,
    config: model_api.ReaLModelConfig,
    head_param_point_to_embedding: bool,
    param_spec: Dict[str, ContiguousParamSpec],
    contiguous_param: torch.Tensor,
    layer_idx_offset: int,
    allocate_only: bool,
):
    for local_layer_idx, l in enumerate(layers):
        layer_idx = local_layer_idx + layer_idx_offset
        for k, v in l.named_parameters():
            spec = param_spec[f"{layer_idx}.{k}"]
            old_param_data = v.data
            target = contiguous_param[spec.start_idx : spec.end_idx].view(spec.shape)
            if not allocate_only:
                target.copy_(old_param_data)
            else:
                if not (
                    head_param_point_to_embedding and layer_idx == config.n_layers + 1
                ):
                    assert old_param_data.shape == torch.Size([0]), (
                        old_param_data.shape,
                        spec.shape,
                        f"{layer_idx}.{k}",
                    )
            recursive_getattr(l, k).data = target



# Copied from the vLLM project: https://github.com/vllm-project/vllm
import dataclasses
from contextlib import contextmanager, nullcontext
from typing import Any, List, Optional

import torch
import torch.distributed
import torch.distributed as dist

import realhf.base.constants as constants
import realhf.base.logging as logging

try:
    import pynvml

    from realhf._C.custom_all_reduce import custom_ar

    @contextmanager
    def _nvml():
        try:
            pynvml.nvmlInit()
            yield
        finally:
            pynvml.nvmlShutdown()

except ImportError:
    # For AMD GPUs
    custom_ar = None
    pynvml = None

    @contextmanager
    def _nvml():
        try:
            yield
        finally:
            pass


logger = logging.getLogger(__name__)

_CA_HANDLE = None
_SUPPORTED_WORLD_SIZES = [2, 4, 6, 8]


@dataclasses.dataclass
class GraphCaptureContext:
    stream: torch.cuda.Stream


@contextmanager
def graph_capture():
    """
    `graph_capture` is a context manager which should surround the code that
    is capturing the CUDA graph. Its main purpose is to ensure that the
    some operations will be run after the graph is captured, before the graph
    is replayed. It returns a `GraphCaptureContext` object which contains the
    necessary data for the graph capture. Currently, it only contains the
    stream that the graph capture is running on. This stream is set to the
    current CUDA stream when the context manager is entered and reset to the
    default stream when the context manager is exited. This is to ensure that
    the graph capture is running on a separate stream from the default stream,
    in order to explicitly distinguish the kernels to capture
    from other kernels possibly launched on background in the default stream.
    """
    stream = torch.cuda.Stream()
    graph_capture_context = GraphCaptureContext(stream)
    maybe_ca_context = nullcontext() if _CA_HANDLE is None else _CA_HANDLE.capture()
    with torch.cuda.stream(stream), maybe_ca_context:
        yield graph_capture_context


@dataclasses.dataclass
class GraphCaptureContext:
    stream: torch.cuda.Stream


@contextmanager
def graph_capture():
    """
    `graph_capture` is a context manager which should surround the code that
    is capturing the CUDA graph. Its main purpose is to ensure that the
    some operations will be run after the graph is captured, before the graph
    is replayed. It returns a `GraphCaptureContext` object which contains the
    necessary data for the graph capture. Currently, it only contains the
    stream that the graph capture is running on. This stream is set to the
    current CUDA stream when the context manager is entered and reset to the
    default stream when the context manager is exited. This is to ensure that
    the graph capture is running on a separate stream from the default stream,
    in order to explicitly distinguish the kernels to capture
    from other kernels possibly launched on background in the default stream.
    """
    stream = torch.cuda.Stream()
    graph_capture_context = GraphCaptureContext(stream)
    maybe_ca_context = nullcontext() if _CA_HANDLE is None else _CA_HANDLE.capture()
    with torch.cuda.stream(stream), maybe_ca_context:
        yield graph_capture_context


def init_custom_ar() -> None:
    global _CA_HANDLE
    if _CA_HANDLE is not None:
        return
    mp_rank = constants.model_parallel_rank()
    mp_world_size = constants.model_parallel_world_size()
    if mp_world_size == 1:
        # No need to initialize custom allreduce for single GPU case.
        return

    full_nvlink = _is_full_nvlink(mp_rank, mp_world_size)
    if mp_world_size > 2 and not full_nvlink:
        logger.warning(
            "Custom allreduce is disabled because it's not supported on"
            " more than two PCIe-only GPUs. "
        )
        return

    if mp_world_size not in _SUPPORTED_WORLD_SIZES:
        logger.warn(
            "Custom allreduce is disabled due to an unsupported world size: "
            "%d. Supported world sizes: %s.",
            mp_world_size,
            str(_SUPPORTED_WORLD_SIZES),
        )
        return

    if not _can_p2p():
        logger.warn(
            "Custom allreduce is disabled because your platform lacks GPU P2P"
            " capability. "
        )
        return

    _CA_HANDLE = CustomAllreduce(mp_rank, mp_world_size, full_nvlink=full_nvlink)


def get_handle() -> Optional["CustomAllreduce"]:
    return _CA_HANDLE


def is_initialized() -> bool:
    return _CA_HANDLE is not None


# query if the set of gpus are fully connected by nvlink (1 hop)
@_nvml()
def _is_full_nvlink(rank, world_size):
    handle = pynvml.nvmlDeviceGetHandleByIndex(rank)
    for i in range(world_size):
        if i != rank:
            try:
                link_state = pynvml.nvmlDeviceGetNvLinkState(handle, i)
                if not link_state:
                    return False
            except pynvml.NVMLError as error:
                logger.info(
                    f'NVLink detection failed with message "{str(error)}". '
                    "This is normal if your machine has no NVLink equipped"
                )
                return False
    return True


def _can_p2p() -> bool:
    parallelism_world_size = constants.parallelism_group_size()
    rank = constants.parallelism_rank()
    assert parallelism_world_size % 2 == 0
    if parallelism_world_size == 1:
        return True
    buf = torch.zeros(1, dtype=torch.int32, device="cuda")
    try:
        if rank % 2 == 0:
            dist.send(
                torch.tensor([1], device="cuda"), constants.to_global_pg_rank(rank + 1)
            )
        else:
            dist.recv(buf, constants.to_global_pg_rank(rank - 1))
    except Exception as e:
        logger.warn(e)
        return False
    return True


class CustomAllreduce:

    _SUPPORTED_WORLD_SIZES = [2, 4, 6, 8]

    # max_size: max supported allreduce size
    def __init__(
        self,
        rank: int,
        world_size: int,
        full_nvlink: bool = True,
        max_size: int = 2**31
        - 2,  # originally 8MB not enough, here is max possible max_size as input
    ) -> None:
        """
        Args:
            group: the process group to work on. If None, it will use the
                default process group.
            device: the device to bind the CustomAllreduce to. If None,
                it will be bind to f"cuda:{local_rank}".
        It is the caller's responsibility to make sure each communicator
        is bind to a unique device, and all communicators in this group
        are in the same node.
        """
        self._IS_CAPTURING = False
        self.group = constants.model_parallel_cpu_group()

        assert (
            dist.get_backend(self.group) != dist.Backend.NCCL
        ), "CustomAllreduce should be attached to a non-NCCL group."

        rank = dist.get_rank(group=self.group)
        world_size = dist.get_world_size(group=self.group)
        self.device = torch.device("cuda")

        # buffers memory are owned by this Python class and passed to C++
        # meta data composes of two parts: meta data for synchronization
        # (256 bytes) and a temporary buffer for storing intermediate
        # allreduce results.
        self.meta = torch.zeros(
            custom_ar.meta_size() + max_size,
            dtype=torch.uint8,
            device=self.device,
        )
        # This is a pre-registered IPC buffer. In eager mode, input tensors
        # are first copied into this buffer before allreduce is performed
        self.buffer = torch.empty(max_size, dtype=torch.uint8, device=self.device)
        # This is a buffer for storing the tuples of pointers pointing to
        # IPC buffers from all ranks. Each registered tuple has size of
        # 8*world_size bytes where world_size is at most 8. Allocating 8MB
        # is enough for 131072 such tuples. The largest model I've seen only
        # needs less than 10000 of registered tuples.
        self.rank_data = torch.empty(
            8 * 1024 * 1024, dtype=torch.uint8, device=self.device
        )
        self.max_size = max_size
        self.rank = rank
        self.world_size = world_size
        handles, offsets = self._get_ipc_meta(self.meta)
        self.full_nvlink = full_nvlink
        self._ptr = custom_ar.init_custom_ar(
            self.meta, self.rank_data, handles, offsets, rank, self.full_nvlink
        )
        self.register_buffer(self.buffer)

    @contextmanager
    def capture(self):
        """
        The main responsibility of this context manager is the
        `register_graph_buffers` call at the end of the context.
        It records all the buffer addresses used in the CUDA graph.
        """
        try:
            self._IS_CAPTURING = True
            yield
        finally:
            self._IS_CAPTURING = False
            self.register_graph_buffers()

    def _get_ipc_meta(self, inp: torch.Tensor):
        data = inp.untyped_storage()._share_cuda_()
        shard_data = (
            data[1],  # ipc handle to base ptr
            data[3],  # offset of base ptr
        )
        return self._gather_ipc_meta(shard_data)

    def _gather_ipc_meta(self, shard_data):
        # Note: don't use `[[None]] * self.world_size` here
        # because it will create a list of the same reference
        all_data: List[Optional[Any]] = [[None] for i in range(self.world_size)]
        all_data[self.rank][0] = shard_data

        ranks = dist.get_process_group_ranks(group=self.group)
        ranks.sort()
        for i, rank in enumerate(ranks):
            dist.broadcast_object_list(
                all_data[i], src=rank, group=self.group, device="cpu"
            )

        # we cannot directly use `dist.all_gather_object` here
        # because it is incompatible with `gloo` backend under inference mode.
        # see https://github.com/pytorch/pytorch/issues/126032 for details.

        handles = []
        offsets = []
        for i in range(len(all_data)):
            handles.append(all_data[i][0][0])  # type: ignore
            offsets.append(all_data[i][0][1])  # type: ignore
        # print(handles, offsets)
        return handles, offsets

    def register_buffer(self, inp: torch.Tensor):
        handles, offsets = self._get_ipc_meta(inp)
        custom_ar.register_buffer(self._ptr, inp, handles, offsets)

    def register_graph_buffers(self):
        handle, offset = custom_ar.get_graph_buffer_ipc_meta(self._ptr)
        handles, offsets = self._gather_ipc_meta((bytes(handle), offset))
        logger.info("Registering %d cuda graph addresses", len(offset))
        custom_ar.register_graph_buffers(self._ptr, handles, offsets)

    def should_custom_ar(self, inp: torch.Tensor):
        should_custom_ar = custom_ar.should_custom_ar(
            inp, self.max_size, self.world_size, self.full_nvlink
        )
        return should_custom_ar

    # all reduce, assuming inp tensor is IPC registered with register_buffer,
    # or, in the context of cuda graphs, register_graph_buffers
    def all_reduce_reg(self, inp: torch.Tensor, out: torch.Tensor = None):
        if out is None:
            out = torch.empty_like(inp)
        custom_ar.all_reduce_reg(self._ptr, inp, out)
        return out

    # all reduce, assuming inp tensor is NOT IPC registered
    def all_reduce_unreg(self, inp: torch.Tensor, out: torch.Tensor = None):
        if out is None:
            out = torch.empty_like(inp)
        custom_ar.all_reduce_unreg(self._ptr, inp, self.buffer, out)
        return out

    def custom_all_reduce(self, input: torch.Tensor) -> Optional[torch.Tensor]:
        # when custom allreduce is disabled, this will be None
        if self._IS_CAPTURING:
            if torch.cuda.is_current_stream_capturing():
                if self.should_custom_ar(input):
                    return self.all_reduce_reg(input)
            else:
                if self.should_custom_ar(input):
                    # if warm up, mimic the allocation pattern
                    # since custom allreduce is out-of-place
                    return torch.empty_like(input)
        else:
            # note: outside of cuda graph context,
            # custom allreduce incurs a cost of cudaMemcpy, which should
            # be small(<=1% of overall latency) compared to the performance
            # gains of using custom kernels
            if self.should_custom_ar(input):
                return self.all_reduce_unreg(input)
        return None

    def close(self):
        if self._ptr:
            custom_ar.dispose(self._ptr)
            self._ptr = 0

    def __del__(self):
        self.close()



import dataclasses
from typing import *

import pytest
import torch
import torch.distributed as dist
import transformers

from realhf.base import constants, logging, testing
from realhf.impl.model.nn.real_llm_api import add_helper_functions

logger = logging.getLogger("tests.test_cpu")


# NOTE: To run test for a new model class, please implement and register `real_config_maker`
# in realhf.api.from_hf.<your_model_class_name> and add the model class name to the
# `model_class` fixture in this file.
@pytest.fixture(params=["llama", "gpt2", "qwen2", "gemma", "mistral"])
def model_class(request):
    return request.param


def maybe_prepare_cpu_env(max_prompt_len: int):
    if not dist.is_initialized():
        # for parametrized runs
        dist.init_process_group(
            "gloo", rank=0, world_size=1, init_method="tcp://localhost:7777"
        )
        import deepspeed

        deepspeed.init_distributed()
        testing.init_global_constants(
            num_dp=1,
            num_mp=1,
            num_pp=1,
            sequence_parallel=False,
            max_prompt_len=max_prompt_len,
        )
        assert dist.get_world_size() == 1, dist.get_world_size()


@pytest.fixture
def mconfig(model_class):
    from realhf.impl.model.nn.real_llm_api import ReaLModel

    mconfig = getattr(ReaLModel, f"make_{model_class}_config")()
    return mconfig


@pytest.fixture
def save_path(tmpdir_factory: pytest.TempdirFactory):
    return tmpdir_factory.mktemp("save_path")


@pytest.fixture
def cpu_real_model(model_class, mconfig, save_path):
    max_prompt_len = 128
    maybe_prepare_cpu_env(max_prompt_len)
    with constants.model_scope(testing.MODEL_NAME):
        from realhf.impl.model.nn.real_llm_api import ReaLModel

        model = ReaLModel(mconfig, dtype=torch.float32, device="cpu")
        add_helper_functions(model)
        model.instantiate()
        model.eval()
        getattr(model, f"to_{model_class}")(None, save_path)
    return model


@pytest.fixture
def cpu_hf_model(save_path):
    hf_model = transformers.AutoModelForCausalLM.from_pretrained(save_path).to(
        torch.float32
    )
    hf_model.eval()
    return hf_model


@torch.no_grad()
def test_inference_cpu_consistency(cpu_real_model, cpu_hf_model, model_class, mconfig):
    max_prompt_len = 128
    with constants.model_scope(testing.MODEL_NAME):
        bs = 10
        torch.manual_seed(1)
        input_ids = torch.randint(
            0, mconfig.vocab_size, (bs, max_prompt_len), dtype=torch.long
        )
        input_lens = torch.full((bs,), max_prompt_len, dtype=torch.int32)
        attention_mask = torch.arange(max_prompt_len)[None, :] < input_lens[:, None]

        logits1 = cpu_hf_model(
            input_ids=input_ids, attention_mask=attention_mask
        ).logits * attention_mask.unsqueeze(-1)
        logits2 = cpu_real_model(
            input_ids=input_ids, attention_mask=attention_mask
        ).logits * attention_mask.unsqueeze(-1)

        assert torch.allclose(logits1, logits2, atol=1e-4), (
            model_class,
            (logits1 - logits2).abs().max(),
        )



import os
import subprocess


def check_ray_availability():
    return (
        int(
            subprocess.run(
                ["ray", "--help"],
                stdout=open(os.devnull, "wb"),
                stderr=open(os.devnull, "wb"),
            ).returncode
        )
        == 0
    )



import dataclasses

from omegaconf import MISSING

from realhf.api.core.dfg import MFCDef, ModelFamily, ModelInterface, ModelInterfaceType
from realhf.api.core.system_api import *
from realhf.api.quickstart.dataset import PromptAnswerDatasetConfig
from realhf.api.quickstart.device_mesh import AllocationConfig
from realhf.api.quickstart.entrypoint import register_quickstart_exp
from realhf.api.quickstart.model import (
    ModelTrainEvalConfig,
    OptimizerConfig,
    get_real_model_config,
)
from realhf.experiments.common.common import CommonExperimentConfig


@dataclasses.dataclass
class SFTConfig(CommonExperimentConfig):
    """SFT experiment configuration.

    It is a subclass of :class:`CommonExperimentConfig`,
    so all CLI options in the base class are available.

    :param model: Model runtime configuration.
    :type model: ModelTrainEvalConfig
    :param allocation: Device allocation and parallelism configuration.
    :type allocation: AllocationConfig
    :param dataset: Dataset configuration
    :type dataset: PromptAnswerDatasetConfig
    """

    model: ModelTrainEvalConfig = dataclasses.field(
        default_factory=ModelTrainEvalConfig
    )
    allocation: AllocationConfig = dataclasses.field(default_factory=AllocationConfig)
    dataset: PromptAnswerDatasetConfig = dataclasses.field(
        default_factory=PromptAnswerDatasetConfig
    )

    @property
    def models(self):
        return {
            "default": self.model,
        }

    @property
    def rpcs(self):
        interface = ModelInterface("sft")
        rpc = MFCDef(
            model_name=ModelName("default", 0),
            interface_type=ModelInterfaceType.TRAIN_STEP,
            interface_impl=interface,
            model_type=self.model.type,
            model_path=self.model.path,
            input_data=["packed_input_ids", "prompt_mask"],
            log_return_value=True,
            min_n_seqs=self.dataset.train_bs_n_seqs,
            max_n_seqs=self.dataset.train_bs_n_seqs,
        )
        return {"default": rpc}

    @property
    def allocations(self):
        return {"default": self.allocation}

    @property
    def datasets(self):
        return [
            Dataset(
                "prompt_answer",
                args=dict(
                    max_length=self.dataset.max_seqlen,
                    dataset_path=self.dataset.train_path,
                ),
            )
        ]

    @property
    def allocations(self):
        return {"default": self.allocation}

    @property
    def eval_datasets(self):
        return [
            Dataset(
                "prompt_answer",
                args=dict(
                    max_length=self.dataset.max_seqlen,
                    dataset_path=self.dataset.valid_path,
                ),
            )
        ]

    @property
    def eval_dataloader(self):
        return DataLoader(
            "packed_eval", args=dict(batch_size=self.dataset.valid_bs_n_seqs)
        )

    @property
    def tokenizer_name_or_path(self):
        return self.model.path


register_quickstart_exp("sft", SFTConfig)



import dataclasses

from omegaconf import MISSING

from realhf.api.core.dfg import MFCDef, ModelInterface, ModelInterfaceType
from realhf.api.core.model_api import GenerationHyperparameters
from realhf.api.core.system_api import *
from realhf.api.quickstart.dataset import PromptOnlyDatasetConfig
from realhf.api.quickstart.device_mesh import AllocationConfig
from realhf.api.quickstart.entrypoint import register_quickstart_exp
from realhf.api.quickstart.model import ModelTrainEvalConfig
from realhf.experiments.common.common import CommonExperimentConfig


@dataclasses.dataclass
class GenerationConfig(CommonExperimentConfig):
    """Generation experiment configuration. Used for testing only.

    It is a subclass of :class:`CommonExperimentConfig`,
    so all CLI options in the base class are available.

    :param model: Model runtime configuration.
    :type model: ModelTrainEvalConfig
    :param gen_params: Generation hyperparameters.
    :type gen_params: GenerationHyperparameters
    :param dataset: Dataset configuration
    :type dataset: PromptOnlyDatasetConfig
    :param allocation: Device allocation and parallelism configuration.
    :type allocation: AllocationConfig
    """

    model: ModelTrainEvalConfig = dataclasses.field(
        default_factory=ModelTrainEvalConfig
    )
    gen: GenerationHyperparameters = dataclasses.field(
        default_factory=GenerationHyperparameters
    )
    dataset: PromptOnlyDatasetConfig = dataclasses.field(
        default_factory=PromptOnlyDatasetConfig
    )
    allocation: AllocationConfig = dataclasses.field(default_factory=AllocationConfig)

    @property
    def models(self):
        return {
            "default": self.model,
        }

    @property
    def rpcs(self):
        interface = ModelInterface("generation", args={"generation_config": self.gen})
        gen = MFCDef(
            model_name=ModelName("default", 0),
            interface_type=ModelInterfaceType.GENERATE,
            model_type=self.model,
            model_path=self.model,
            interface_impl=interface,
            input_data=["packed_prompts"],
            balanced_dp=True,
            log_return_value=True,
            min_n_seqs=self.dataset.train_bs_n_seqs,
            max_n_seqs=self.dataset.train_bs_n_seqs,
        )
        return {"default": gen}

    @property
    def allocations(self):
        return {"default": self.allocation}

    @property
    def max_prompt_len(self):
        return self.dataset.max_prompt_len

    @property
    def datasets(self):
        return [
            Dataset(
                "prompt",
                args=dict(
                    dataset_path=self.dataset.path,
                    max_length=self.dataset.max_prompt_len,
                    pad_to_max_length=self.dataset.pad_to_max_length,
                ),
            )
        ]

    @property
    def allocations(self):
        return {"default": self.allocation}

    @property
    def tokenizer_name_or_path(self):
        return self.model.path

    @property
    def exp_ctrl(self):
        return ExperimentSaveEvalControl(
            total_train_epochs=1,
            save_freq_steps=None,
            eval_freq_epochs=None,
        )


register_quickstart_exp("gen", GenerationConfig)



import contextlib
import dataclasses
import enum
import os
import pickle
import time
from collections import defaultdict
from statistics import mean
from typing import TYPE_CHECKING, Callable, List, Optional, Union

import numpy as np
import psutil
import pynvml

import realhf.base.constants as constants
import realhf.base.logging as logging

if TYPE_CHECKING:
    from realhf.api.core.config import ModelName

logger = logging.getLogger("benchmark")

IF_MARK = False


def process_memory_mb(name):
    process = psutil.Process()
    memory_info = process.memory_info()
    memory_usage_mb = memory_info.rss / 1024**2
    pid = process.pid
    logger.debug(f"Process PID {pid} memory usage @{name}: {memory_usage_mb}.")


def gpu_memory_mb(name):
    import torch.distributed as dist
    from deepspeed.accelerator import get_accelerator

    logger.debug(
        f"{name} GPU rank {dist.get_rank()}: memory usage: {round(get_accelerator().memory_allocated() / 1024**2, 2)}MB, "
        f"max memory usage: {round(get_accelerator().max_memory_allocated() / 1024**2, 2)}MB"
    )


def mock_time_mark(name, identifier, t, step):
    if IF_MARK:
        logger.debug(f"*{name}* #{identifier}#  ${t}$ ns step &{step}&")


def time_mark(name, identifier, step=0):
    if IF_MARK:
        logger.debug(
            f"*{name}* #{identifier}#  ${int(time.time_ns())}$ ns step &{step}&"
        )


def parse_time_mark_in_line(line, name, step_range=None):
    if f"*{name}*" in line:
        identifer, t, step = (
            line.split("#")[1],
            int(line.split("$")[1]),
            int(line.split("&")[1]),
        )
        if step_range:
            if step >= step_range[1] or step < step_range[0]:
                return None
        return identifer, t

    else:
        return None


def parse_time_mark_in_file(file, name, step_range=None):
    time_points = defaultdict(list)
    with open(file, "r") as f:
        count = 0
        res_count = 0
        for line in f.readlines():
            count += 1
            res = parse_time_mark_in_line(line, name, step_range=step_range)
            if res is not None:
                res_count += 1
                identifier, time_point = res
                time_points[identifier].append(time_point)
        # print(f"file {file} name {name} line count {count} res count {res_count}")
    return time_points


def parse_time_mark_in_dir(dir, name, step_range=None):
    time_points = {}
    for file in os.listdir(dir):
        file_path = os.path.join(dir, file)
        tpsf = parse_time_mark_in_file(file_path, name, step_range=step_range)
        for k, v in tpsf.items():
            if k not in time_points:
                time_points[k] = v
            else:
                time_points[k].extend(v)
    return time_points


MATPLOTLIB_COLORS = [
    "red",
    "blue",
    "green",
    "yellow",
    "orange",
    "purple",
    "pink",
    "black",
    "brown",
    "gray",
    "cyan",
    "magenta",
    "lime",
    "olive",
    "navy",
]


def summary_time_points(
    start_keys,
    end_keys,
    identifiers,
    dir_name=None,
    file_name=None,
    start_time=None,
    figsize=(12, 4),
    end_time=None,
    step_range=None,
    save_fig_path="time_points.png",
    draw_boundary=False,
):
    """Plot and summary time marks in logs"""
    import matplotlib.pyplot as plt

    assert file_name or dir_name, "dir or file name must be specified"
    all_time_points = {}
    if file_name is None:
        for k in start_keys:
            all_time_points[k] = parse_time_mark_in_dir(
                dir_name, k, step_range=step_range
            )
        for k in end_keys:
            all_time_points[k] = parse_time_mark_in_dir(
                dir_name, k, step_range=step_range
            )
    else:
        for k in start_keys:
            all_time_points[k] = parse_time_mark_in_file(
                file_name, k, step_range=step_range
            )
        for k in end_keys:
            all_time_points[k] = parse_time_mark_in_file(
                file_name, k, step_range=step_range
            )

    fig, ax = plt.subplots(1, 1, figsize=figsize)
    ax.set_ylim(-1, len(identifiers))
    ax.set_yticks(list(range(len(identifiers))))
    ax.set_yticklabels(identifiers)

    label_set = {sk: False for sk in start_keys}
    infos = {}
    min_time = None
    max_time = None
    for id_index, identifier in enumerate(identifiers):
        time_sum = {}
        time_list = {}
        for start_key_idx, (start_key, end_key) in enumerate(zip(start_keys, end_keys)):
            # print(start_key, identifier, all_time_points[start_key])
            time_sum[start_key] = 0
            time_list[start_key] = []
            try:
                start_time_points = np.array(all_time_points[start_key][identifier])
                end_time_points = np.array(all_time_points[end_key][identifier])
            except KeyError:
                continue
            assert len(start_time_points) == len(end_time_points)

            if start_time is not None:
                valid_indices_st = np.where(start_time_points > start_time)
                valid_indices_et = np.where(start_time_points < end_time)
                valid_indices = np.intersect1d(valid_indices_st, valid_indices_et)
                start_time_points = start_time_points[valid_indices]
                end_time_points = end_time_points[valid_indices]

            # print(id_index, identifier, start_key_idx, start_key, end_key, start_time_points, end_time_points)

            # plot time point pairs
            for stp, etp in zip(list(start_time_points), list(end_time_points)):
                min_time = stp if min_time is None else min(min_time, stp)
                max_time = etp if max_time is None else max(max_time, etp)
                time_sum[start_key] += etp - stp
                time_list[start_key].append(etp - stp)

                if label_set[start_key] is False:
                    label = start_key
                    label_set[start_key] = True
                else:
                    label = None

                # print(f"id={identifier} start_key={start_key} left={stp%1000} width={etp-stp}")
                # print((etp-stp)//1e6)
                ax.barh(
                    y=id_index,
                    width=etp - stp,
                    left=stp,
                    height=0.8,
                    color=MATPLOTLIB_COLORS[start_key_idx],
                    label=label,
                )

                if draw_boundary:
                    ax.plot(
                        [stp, stp],
                        [id_index - 0.4, id_index + 0.4],
                        color="black",
                        linestyle="-",
                        linewidth=0.5,
                    )
                    ax.plot(
                        [etp, etp],
                        [id_index - 0.4, id_index + 0.4],
                        color="black",
                        linestyle="-",
                        linewidth=0.5,
                    )

        infos[identifier] = (time_sum, time_list)

    ax.set_xlim(min_time, max_time)
    total_width = max_time - min_time
    xticks = np.arange(
        min_time - total_width // 12, max_time - total_width // 12, 10 * 1e9
    )
    xtick_labels = [f"{int((i//1e9)%1000)}" for i in xticks]
    ax.set_xticks(xticks)
    ax.set_xticklabels(xtick_labels)

    # summary time cost percent
    for id_index, identifier in enumerate(identifiers):
        print("=" * 30)
        print(f"Identifier {identifier} time cost percent:")
        bubble_time = 100
        time_sum, time_list = infos[identifier]
        for k in time_sum:
            time_perc = round(time_sum[k] / (max_time - min_time) * 100, 2)
            # print time cost percent
            avg_val = (
                round(mean(time_list[k]) / 10e6, 2) if len(time_list[k]) > 0 else "-"
            )
            max_val = (
                round(max(time_list[k]) / 10e6, 2) if len(time_list[k]) > 0 else "-"
            )
            min_val = (
                round(min(time_list[k]) / 10e6, 2) if len(time_list[k]) > 0 else "-"
            )

            bubble_time -= time_perc
            print(
                f"{k} -- {time_perc} %, "
                f"avg, min, max = {avg_val}, {min_val}, {max_val} ms, "
                f"sum, n = {round(time_sum[k]/10e6, 2)} ms, {len(time_list[k])}"
            )
        print(f"bubble time -- {round(bubble_time, 2)}%")

    plt.legend(loc=(1.01, 0.0))
    plt.tight_layout()

    plt.savefig(save_fig_path)


def gpu_utilization_monitor(worker_idx: int, interval: float, ttl: float):
    pynvml.nvmlInit()
    gpu_idx = worker_idx % 8
    tik = time.time()
    while time.time() - tik < ttl:
        handle = pynvml.nvmlDeviceGetHandleByIndex(gpu_idx)
        utilization = pynvml.nvmlDeviceGetUtilizationRates(handle)
        memory_info = pynvml.nvmlDeviceGetMemoryInfo(handle)
        total_memory = memory_info.total / (1024**2)  # Convert bytes to megabytes
        used_memory = memory_info.used / (1024**2)
        memory_usage_percentage = (used_memory / total_memory) * 100
        logger.debug(
            f"Worker Index {worker_idx}, GPU {gpu_idx}: "
            f"Compute Utilization - {utilization.gpu}%, "
            f"Total Memory - {total_memory:.2f}MB, Used Memory - {used_memory:.2f}MB, "
            f"Memory Usage - {memory_usage_percentage:.2f}%"
        )
        time.sleep(interval)
    pynvml.nvmlShutdown()


# Helper function to calculate FLOPs using the Megatron-LM paper's formula
def calculate_llama_train_flops(
    checkpoint_activations_factor: int,
    batch_size: int,
    seqlens: List[int],
    num_layers: int,
    hidden_size: int,
    intermediate_size: int,
    vocab_size: int,
):
    return checkpoint_activations_factor * caculuate_llama_forward_flops(
        batch_size,
        seqlens,
        num_layers,
        hidden_size,
        intermediate_size,
        vocab_size,
    )


def caculuate_llama_forward_flops(
    batch_size: int,
    seqlens: List[int],
    num_layers: int,
    hidden_size: int,
    intermediate_size: int,
    vocab_size: int,
):
    assert len(seqlens) == batch_size
    attn_flops = sum(x**2 for x in seqlens) * hidden_size
    return (
        2
        * num_layers
        * (
            4 * sum(seqlens) * hidden_size**2
            + 2 * attn_flops
            + 3 * sum(seqlens) * hidden_size * intermediate_size
        )
        + 4 * sum(seqlens) * vocab_size * hidden_size
    )


def calculate_llama_gen_flops(
    batch_size,
    prompt_lens,
    gen_len,
    num_layers,
    hidden_size,
    intermediate_size,
    vocab_size,
):
    flops = caculuate_llama_forward_flops(
        batch_size,
        prompt_lens,
        num_layers,
        hidden_size,
        intermediate_size,
        vocab_size,
    )
    for i in range(gen_len):
        prefix_lens = [x + i for x in prompt_lens]
        flops += (
            2
            * num_layers
            * (
                4 * batch_size * hidden_size**2
                + 2 * (sum(prefix_lens) + batch_size) * hidden_size
                + 3 * batch_size * hidden_size * intermediate_size
            )
            + 4 * batch_size * vocab_size * hidden_size
        )
    return flops


class CUDATimeMarkType(enum.Enum):
    forward = "forward"
    backward = "backward"
    optim_step = "optim_step"
    comm = "comm"
    misc = "misc"
    mem_layout = "memory_layout"


@dataclasses.dataclass
class TimeMarkEntry:
    name: str
    model_name: "ModelName"
    type_: CUDATimeMarkType
    start_time: int
    end_time: int


TIME_MARK_DB = []


def cuda_tmark(name: str, type_: CUDATimeMarkType):
    if os.getenv("REAL_CUDA_TMARK", None) == "1":

        def wrapper(f: Callable):

            def _wrapped_f(*args, **kwargs):
                import torch

                if torch.cuda.is_available():
                    from realhf.base.constants import _model_name

                    torch.cuda.synchronize()
                    tik = time.time_ns()
                    res = f(*args, **kwargs)
                    torch.cuda.synchronize()
                    tok = time.time_ns()
                    global TIME_MARK_DB
                    TIME_MARK_DB.append(
                        TimeMarkEntry(name, _model_name, type_, tik, tok)
                    )
                else:
                    res = f(*args, **kwargs)
                return res

            return _wrapped_f

    else:

        def wrapper(f):
            return f

    return wrapper


@contextlib.contextmanager
def cuda_tmarked(name: str, type_: CUDATimeMarkType):
    if os.getenv("REAL_CUDA_TMARK", None) == "1":
        import torch

        if torch.cuda.is_available():
            from realhf.base.constants import _model_name

            torch.cuda.synchronize()
            tik = time.time_ns()
    yield
    if os.getenv("REAL_CUDA_TMARK", None) == "1":
        if torch.cuda.is_available():
            torch.cuda.synchronize()
            tok = time.time_ns()
            global TIME_MARK_DB
            TIME_MARK_DB.append(TimeMarkEntry(name, _model_name, type_, tik, tok))


def fetch_latest_tmark():
    global TIME_MARK_DB
    return TIME_MARK_DB[-1]


def dump_tmark_db(worker_idx):
    if os.getenv("REAL_CUDA_TMARK", None) != "1":
        return
    fn = os.path.join(
        constants.LOG_ROOT,
        constants.experiment_name(),
        constants.trial_name(),
        f"time_marks{worker_idx}.pkl",
    )
    global TIME_MARK_DB
    with open(fn, "wb") as f:
        pickle.dump(TIME_MARK_DB, f)
    TIME_MARK_DB.clear()


COMPUTE_KERNEL_KEYS = [
    "elementwise_kernel",
    "gemm_",
    "aten::",
    "at::native::",
    "flash",
    "backward_kernel",
    "reduce_kernel",
    "multi_tensor_apply",
]

COMM_KERNEL_KEYS = [
    "c10d::",
    "nccl::",
    "ncclDevKernel",
]

MEM_KERNEL_KEYS = [
    "Memcpy",
    "Memset",
]

MISC_KERNEL_KEYS = [
    "at_cuda_detail",
    "CudaCodeGen",
]


@dataclasses.dataclass
class CUDAKernelTime:  # in us
    compute: int
    comm: int
    mem: int
    misc: int

    @classmethod
    def from_profiler(cls, p):
        import torch

        compute_time = comm_time = mem_time = misc_time = 0
        unknown_keys = []
        for x in p.key_averages():
            if x.device_type != torch.autograd.DeviceType.CUDA:
                continue
            if x.self_cuda_time_total <= 0:
                continue
            if "waitevent" in x.key.lower():
                print(x.key)
            if any(k in x.key for k in COMPUTE_KERNEL_KEYS):
                compute_time += x.self_cuda_time_total
            elif any(k in x.key for k in COMM_KERNEL_KEYS):
                comm_time += x.self_cuda_time_total
            elif any(k in x.key for k in MEM_KERNEL_KEYS):
                mem_time += x.self_cuda_time_total
            elif any(k in x.key for k in MISC_KERNEL_KEYS):
                misc_time += x.self_cuda_time_total
            else:
                unknown_keys.append(x)
        if unknown_keys:
            raise NotImplementedError(
                f"Unknown keys: {[(x.key, x.self_cuda_time_total) for x in unknown_keys]}"
            )
        return cls(compute=compute_time, comm=comm_time, mem=mem_time, misc=misc_time)

    def __add__(self, other):
        return CUDAKernelTime(
            compute=self.compute + other.compute,
            comm=self.comm + other.comm,
            mem=self.mem + other.mem,
            misc=self.misc + other.misc,
        )

    @property
    def total_secs(self):
        return (self.compute + self.comm + self.mem + self.misc) / 1e6

    def __truediv__(self, x):
        return CUDAKernelTime(
            compute=self.compute / x,
            comm=self.comm / x,
            mem=self.mem / x,
            misc=self.misc / x,
        )

    def __repr__(self):
        return f"CUDAKernelTime(compute={self.compute}us, comm={self.comm}us, mem={self.mem}us, misc={self.misc}us)"
